import copy
import hashlib
import json
import re
import time
from logging import error, info, warning
from typing import Any, Iterable, Union

import act.api
from act.api.re import UUID_MATCH

from . import DEFAULT_FACT_VALIDATOR
from .base import (ActBase, Comment, NameSpace, Organization, Origin,
                   UserReference, origin_serializer)
from .obj import Object, ObjectType
from .schema import Field, MissingField, ValidationError, schema_doc


class IllegalFactChain(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class UnknownType(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class RelevantObjectBindings(ActBase):
    SCHEMA = [
        Field(
            "source_object_type",
            deserializer=ObjectType,
            serializer=lambda source_object_type: source_object_type.id,
        ),
        Field(
            "destination_object_type",
            deserializer=ObjectType,
            serializer=lambda destination_object_type: destination_object_type.id,
        ),
        Field("bidirectional_binding", default=False),
    ]

    def __hash__(self):
        return hash(
            (
                self.__class__.__name__,
                self.source_object_type,
                self.destination_object_type,
                self.bidirectional_binding,
            )
        )


class RelevantFactBindings(ActBase):
    """Meta Fact Type"""

    SCHEMA = [
        Field("name"),
        Field("id"),
        Field("fact_type", flatten=True),
    ]

    def __hash__(self):
        return hash((self.__class__.__name__, self.name))

    def serialize(self):
        # Return None for empty objects (non initialized objects)
        if not self.id:
            return None

        return {"factType": self.id}

    def __bool__(self):
        # Return None for empty objects (non initialized objects)
        if not self.id:
            return False

        return True


class FactType(ActBase):
    """Manage FactType"""

    SCHEMA = [
        Field("name"),
        Field("id"),
        Field("default_confidence"),
        Field("validator", default="RegexValidator"),
        Field("validator_parameter", default=DEFAULT_FACT_VALIDATOR),
        Field("relevant_object_bindings", deserializer=RelevantObjectBindings),
        Field("relevant_fact_bindings", deserializer=RelevantFactBindings),
        Field("namespace", deserializer=NameSpace),
    ]

    @schema_doc(SCHEMA)
    def __init__(self, *args, **kwargs):
        super(FactType, self).__init__(*args, **kwargs)

    def add(self):
        params = self.serialize()

        fact_type = self.api_post("v1/factType", **params)["data"]

        # Empty data and load new result from response
        self.data = {}
        self.deserialize(**fact_type)

        info("Created fact type: {}".format(self.name))

        return self

    def add_object_binding(
        self, source_object_type, destination_object_type, bidirectional_binding=False
    ):
        """Add bindings
        Args:
            source_object_type (objectType):        Source Object Type
            destination_object_type (objectType):   Destination Object Type
            bidirectional_binding (boolean):        Whether the binding is bidirectional"""
        return self.add_object_bindings(
            [
                RelevantObjectBindings(
                    source_object_type, destination_object_type, bidirectional_binding
                )
            ]
        )

    def add_object_bindings(self, add_bindings):
        """Add multiple object bindings
        Args:
            add_bindings (relevantObjectBindings[]): List of bindings which must be a
                                                     tuple of RelevantObjectBindings"""
        if not self.id:
            raise MissingField("Must have fact type ID")

        url = "v1/factType/uuid/{}".format(self.id)

        existing_bindings = set(self.relevant_object_bindings)

        # Exclude already existing bindings
        new_bindings = [
            binding for binding in add_bindings if binding not in existing_bindings
        ]

        if not new_bindings:
            warning("All bindings specified for {} already exists".format(self.name))
            return self

        serialized_bindings = [binding.serialize() for binding in new_bindings]

        fact_type = self.api_put(url, addObjectBindings=serialized_bindings)["data"]

        self.data = {}
        self.deserialize(**fact_type)

        # Emit log of created bindings
        for binding in new_bindings:
            info(
                "Added bindings to fact type {}: {} -> {} (bidirectional_binding={})".format(
                    self.name,
                    binding.source_object_type.data.get("name", None),
                    binding.destination_object_type.data.get("name", None),
                    binding.bidirectional_binding,
                )
            )

        return self

    def add_fact_binding(self, fact_type):
        """Add bindings
        Args:
            fact_type (factType):   Fact type"""
        return self.add_fact_bindings(
            [RelevantFactBindings(name=fact_type.name, id=fact_type.id)]
        )

    def add_fact_bindings(self, add_bindings):
        """Add multiple fact bindings
        Args:
            add_bindings (relevantFactBindings[]): List of Fact bindings"""
        if not self.id:
            raise MissingField("Must have fact type ID")

        url = "v1/factType/uuid/{}".format(self.id)

        existing_bindings = set(self.relevant_fact_bindings)

        # Exclude already existing bindings
        new_bindings = [
            binding for binding in add_bindings if binding not in existing_bindings
        ]

        if not new_bindings:
            warning("All bindings specified for {} already exists".format(self.name))
            return self

        serialized_bindings = [binding.serialize() for binding in new_bindings]

        fact_type = self.api_put(url, addFactBindings=serialized_bindings)["data"]

        self.data = {}
        self.deserialize(**fact_type)

        # Emit log of created bindings
        for binding in new_bindings:
            info(
                "Added bindings to meta fact type {}: {}".format(
                    self.name, binding.name
                )
            )
        return self

    def rename(self, name):
        if not self.id:
            raise MissingField("Must have fact type ID")

        old_name = self.name

        url = "v1/factType/uuid/{}".format(self.id)
        fact_type = self.api_put(url, name=name)["data"]

        self.data = {}
        self.deserialize(**fact_type)

        info("Renamed fact type {}: {} -> {}".format(self.id, old_name, name))

        return self

    def __hash__(self):
        return hash(
            (
                self.name,
                self.namespace,
            )
        )


def object_serializer(obj):
    if not obj:
        return None

    if obj.id is None and obj.type.name is None and obj.value is None:
        return None

    if obj.id:
        return obj.id

    return "{}/{}".format(obj.type.name, obj.value)


class AbstractFact(ActBase):

    SCHEMA = [
        Field(
            "type", deserializer=FactType, serializer=lambda fact_type: fact_type.name
        ),
        Field("value", default=None),
        Field("id", serializer=False),
        Field("flags", serializer=False),
        Field("origin", deserializer=Origin),
        Field("comment"),
        Field("added_by", deserializer=UserReference, serializer=False),
        Field("trust", serializer=False),
        Field("confidence"),
        Field("certainty", serializer=False),
        Field("timestamp", serializer=False),
        Field("last_seen_timestamp", serializer=False),
        Field("organization", deserializer=Organization),
        Field("access_mode"),
        Field("source_object", deserializer=Object),
        Field("destination_object", deserializer=Object),
        Field("bidirectional_binding", default=False),
        Field("last_seen_by", deserializer=UserReference, serializer=False),
    ]

    @schema_doc(SCHEMA)
    def __init__(self, *args, **kwargs):
        super(AbstractFact, self).__init__(*args, **kwargs)

    def get(self):
        """Get fact"""

        if not self.id:
            raise MissingField("Must have fact ID to get, or use search instead")

        fact = self.api_get("v1/fact/uuid/{}".format(self.id))["data"]
        self.data = {}
        self.deserialize(**fact)
        return self

    def set_defaults(self):
        """
        Set fact defaults from config if we have not specified
        - access_mode
        - origin
        - organization
        """

        self.access_mode = self.access_mode or self.config.access_mode
        self.organization = self.organization or self.config.organization

        if self.access_mode not in act.api.ACCESS_MODES:
            raise act.api.base.ArgumentError(
                f"Illegal access_mode: {self.access_mode}. Must be one of "
                + f"{','.join(act.api.ACCESS_MODES)}"
            )

        if not self.origin:
            # If origin is not specified explicit on the fact, use origin from default config
            if self.config.origin_id:
                self.origin = Origin(id=self.config.origin_id)
            elif self.config.origin_name:
                self.origin = Origin(name=self.config.origin_name)

        return self

    # pylint: disable=unused-argument,dangerous-default-value

    def get_acl(self):
        """Get acl"""

        if not self.id:
            raise MissingField("Must have fact ID to resolve ACL")

        return self.api_get("v1/fact/uuid/{}/access".format(self.id))["data"]

    def grant_access(self, subject_uuid):
        """Grant access - not implemented yet"""

        raise act.api.base.NotImplemented(
            "Grant access is not implemented, ignoring {}".format(subject_uuid)
        )

    def get_comments(self):
        """Get comments
        Args:
            None

        Returns ActResultSet of Comments.

        """

        if not self.id:
            raise MissingField("Must have fact ID to get comments")

        res = self.api_get("v1/fact/uuid/{}/comments".format(self.id))
        return act.api.base.ActResultSet(res, Comment)

    def add_comment(self, comment, reply_to=None):
        """Add comment
        Args:
            comment (str):    Comment to add
            reply_to (uuid):  Comment is a reply to another comment (specified by uuid)

        Returns Fact object
        """

        if not self.id:
            raise MissingField("Must have fact ID to add comment")

        if reply_to and not re.search(UUID_MATCH, reply_to):
            raise ValidationError("reply_to is not a valid UUID: {}".format(reply_to))

        self.api_post(
            "v1/fact/uuid/{}/comments".format(self.id),
            comment=comment,
            replyTo=reply_to,
        )

        return self

    def retract(
        self, organization=None, source=None, access_mode=None, comment=None, acl=[]
    ):
        """Retract fact
        Args:
            organization (str): Set owner of new Fact. If not set the current user's
                                organization will be used (takes Organization UUID)
            source (str):       Set Source of new Fact. If not set the current user
                                will be used as Source (takes Source UUID)
            access_mode:        Set access mode of new Fact. If not set the accessMode
                                from the retracted Fact will be used = ['Public',
                                'RoleBased', 'Explicit']
            comment (str):      If set adds a comment to new Fact
            acl (str[] | str):  If set defines explicitly who has access to new Fact (takes Subject UUIDs)

        All arguments are optional.

        Returns retracted fact.
        """

        params = act.api.utils.prepare_params(locals(), ensure_list=["acl"])

        if self.id:
            url = "v1/fact/uuid/{}/retract".format(self.id)
        else:
            raise MissingField("Must have object ID to retract object")

        fact = self.api_post(url, **params)["data"]

        meta = MetaFact(**fact)
        # Add config to meta fact (user/auth)
        meta.configure(self.config)
        meta.set_defaults()

        return meta

    def __str__(self):
        """
        Construnct string representation on this format
        (src_obj_type/src_obj_value) -[fact_type/fact_value]-> (dest_obj_type/dest_obj_value)
        """

        out = ""

        # Include source object if set
        if self.source_object:
            out += "({}/{}) -".format(
                self.source_object.type.name, self.source_object.value
            )

        # Add fact type
        if self.type:
            out += "[{}".format(self.type.name)
        else:
            error("Fact has no type, %s", self.data)

        # Add value if set
        if self.value and not self.value.startswith("-"):
            out += "/{}".format(self.value)
        out += "]"

        # Add destination object if set
        if self.destination_object:

            # Use arrow unless bidirectional
            if self.bidirectional_binding:
                out += "-"
            else:
                out += "->"

            out += " ({}/{})".format(
                self.destination_object.type.name, self.destination_object.value
            )

        return out


class Fact(AbstractFact):
    """Manage facts"""

    SCHEMA = AbstractFact.SCHEMA + [
        Field("in_reference_to", serializer=False),
    ]

    def __hash__(self):
        """
        Hash of the fact. Include all fields that makes the fact unique.
        """

        return hash(
            (
                self.__class__.__name__,
                self.type,
                self.value,
                self.origin,
                self.confidence,
                self.organization,
                self.access_mode,
                self.source_object,
                self.destination_object,
                self.bidirectional_binding,
            )
        )

    def source(self, *args, **kwargs):
        """Add source binding. Accepts the same arguments as Object()"""

        self.data["source_object"] = Object(*args, **kwargs)

        if not self.source_object.id and not (
            self.source_object.type and self.source_object.value
        ):
            raise MissingField(
                f"Must have either object_id or object_type and object_value: {self.data}"
            )

        return self

    def destination(self, *args, **kwargs):
        """Add source binding. Accepts the same arguments as Object()"""

        self.data["destination_object"] = Object(*args, **kwargs)

        if not self.destination_object.id and not (
            self.destination_object.type and self.destination_object.value
        ):
            raise MissingField(
                f"Must have either object_id or object_type and object_value: {self.data}"
            )

        return self

    def format_objects(self, object_formatter=None):
        if not object_formatter:
            if not self.config or not self.config.object_formatter:
                raise act.api.base.ArgumentError(
                    "object_formatter must be specified as argument or in configuration"
                )

            object_formatter = self.config.object_formatter

        if self.source_object:
            self.source_object.value = object_formatter(
                self.source_object.type.name, self.source_object.value
            )
        if self.destination_object:
            self.destination_object.value = object_formatter(
                self.destination_object.type.name,
                self.destination_object.value,
            )

        return self

    def validate_and_raise(self, object_validator=None):

        if not object_validator:
            if not self.config or not self.config.object_validator:
                raise act.api.base.ArgumentError(
                    "object_validator must be specified as argument or in configuration"
                )

            object_validator = self.config.object_validator

        if self.source_object == self.destination_object:
            raise act.api.base.ValidationError(
                f"Source object can not be equal to destination object: {self.json()}"
            )

        if self.source_object and not object_validator(
            self.source_object.type.name, self.source_object.value
        ):
            raise act.api.base.ValidationError(
                f"Source object does not validate: {self.json()}"
            )

        if self.destination_object and not object_validator(
            self.destination_object.type.name, self.destination_object.value
        ):
            raise act.api.base.ValidationError(
                f"Destination object does not validate: {self.json()}"
            )

        return True

    def bidirectional(
        self,
        source_object_type,
        source_object_value,
        destination_object_type,
        destination_object_value,
    ):
        """Add bidirectional binding."""

        self.data["source_object"] = Object(source_object_type, source_object_value)
        self.data["destination_object"] = Object(
            destination_object_type, destination_object_value
        )
        self.data["bidirectional_binding"] = True

        return self

    def add(self):
        """Add fact"""
        started = time.time()

        # Construct parameters to be sent to backend when creating fact
        params = {
            k: v
            for k, v in self.serialize().items()
            # Exclude inReferenceTo, which is added automatically for retracted facts and meta facts
            if k not in ("inReferenceTo")
        }

        # Special serializer for source/destination objects
        params["sourceObject"] = object_serializer(self.source_object)
        params["destinationObject"] = object_serializer(self.destination_object)
        params["origin"] = origin_serializer(self.origin)

        # Replace "" in value parameter with None
        # This is for backward compatibilty for output produced on act-api <=v2.0.3
        # Can be removed when we bump major version to 3.x.x
        if params.get("value") == "":
            warning(
                "Replaced empty string in fact value with None. "
                "This should be fixed in workers producing these facts, %s",
                params,
            )
            params["value"] = None

        fact = self.api_post("v1/fact", **params)["data"]

        # Empty data and load new result from response
        self.data = {}
        self.deserialize(**fact)

        info(
            "Created fact in %.2fs: data=%s" % (time.time() - started, json.dumps(fact))
        )

        return self

    # pylint: disable=unused-argument,dangerous-default-value
    def meta(self, *args, **kwargs):
        """Create meta fact
        Creates a new meta fact with reference to this fact.
        Takes the same arguments as Fact(), except for in_reference_to.
        Returns meta fact
        """

        # Use all values used in serialization as parameters
        # to create a referenced object
        ref_arg = self.serialize()

        # ID is not included in serialization, so this must be explicitly added
        if self.id:
            ref_arg["id"] = self.id

        ref = Fact(**ref_arg)

        meta = MetaFact(*args, in_reference_to=ref, **kwargs)

        # Add config to meta fact (user/auth)
        meta.configure(self.config)
        meta.set_defaults()

        return meta

    # pylint: disable=unused-argument,dangerous-default-value
    def get_meta(self, before=None, after=None, limit=None):
        """Get meta facts
        Args:
            before (timestamp):           Only return Facts added before a specific
                                          timestamp. Timestamp is on this format:
                                          2016-09-28T21:26:22Z
            after (timestamp):            Only return Facts added after a specific
                                          timestamp. Timestamp is on this format:
                                          2016-09-28T21:26:22Z
            limit (integer):              Limit the number of returned Objects
                                          (default 25, 0 means all)
        """

        params = act.api.utils.prepare_params(locals())

        if not self.id:
            raise MissingField("Must have fact ID to get comments")

        res = self.api_get("v1/fact/uuid/{}/meta".format(self.id), params=params)
        return act.api.base.ActResultSet(res, MetaFact, config=self.config)


class MetaFact(AbstractFact):
    """Manage meta facts"""

    SCHEMA = AbstractFact.SCHEMA + [
        Field("in_reference_to", deserializer=Fact),
    ]

    def __hash__(self):
        """
        Hash of the fact. Include all fields that makes the fact unique.
        """

        return hash(
            (
                self.__class__.__name__,
                self.type,
                self.value,
                self.origin,
                self.confidence,
                self.organization,
                self.access_mode,
                self.in_reference_to,
            )
        )

    def add(self):
        """Add meta fact to platform"""

        started = time.time()
        if not self.in_reference_to:
            raise MissingField("Fact is not a meta fact (in_reference_to is not set)")

        if not self.in_reference_to.id:
            raise MissingField("Referenced fact must have fact ID")

        params = {
            k: v
            for k, v in self.serialize().items()
            if k not in ("inReferenceTo", "bidirectionalBinding") and v
        }

        # Special serializer for origin
        params["origin"] = origin_serializer(self.origin)

        url = "v1/fact/uuid/{}/meta".format(self.in_reference_to.id)

        meta_fact = self.api_post(url, **params)["data"]

        self.data = {}
        self.deserialize(**meta_fact)
        info(
            "Created meta fact in %.2fs: data=%s"
            % (time.time() - started, json.dumps(meta_fact))
        )

        return self


def auto_fact_type(**kwargs: Any) -> Union[MetaFact, Fact]:
    """Guess type based on keys in dictionary and return Object"""
    if kwargs.get("inReferenceTo"):
        return MetaFact(**kwargs)
    elif any([kwargs.get("sourceObject"), kwargs.get("destinationObject")]):
        return Fact(**kwargs)
    else:
        raise UnknownType("Unable to guess element type: {}".format(kwargs))


def fact_chain_seed(*facts):
    """
        Get seed of the fact chain. The seed is the string representations of all facts sorted
        and joined with newline.

    Args:
        *facts [Facts]: ACT facts

    Returns: string representation of the seed
    """

    known_objects = [
        fact.source_object for fact in facts if fact.source_object.value != "*"
    ] + [
        fact.destination_object
        for fact in facts
        if fact.destination_object.value != "*"
    ]

    if len(known_objects) != 2:
        raise IllegalFactChain(
            "There should be exactly two known objects in a fact chain: {}".format(
                facts
            )
        )

    return "\n".join(sorted([str(fact) for fact in facts]))


def fact_chain(*facts: Fact) -> Iterable[Fact]:
    """
        Return fact chain of all facts.
        A fact chain is a modified version of the facts, where all placeholder
        object values are replace with a hash using the all incoming/outgoing
        links from the placeholder as a seed

    Args:
        *facts [Facts]: ACT facts

    Returns: facts [Facts]: ACT facts
    """

    # Calculate the hash for this chain, based on the seed
    sha256sum = hashlib.sha256(fact_chain_seed(*facts).encode("utf8")).hexdigest()

    chain: Iterable[Fact] = copy.copy(facts)

    # Return the fact chain, with all object values replaced "[placeholder[<HASH>]]"
    for fact in chain:
        for obj in (fact.source_object, fact.destination_object):
            if obj.value == "*":
                obj.data["value"] = "[placeholder[{}]]".format(sha256sum)

    return chain



import re

import pytest
import responses
from act_test import get_mock_data

import act
from act.api.re import UUID, UUID_MATCH


# pylint: disable=no-member
@responses.activate
def test_get_object_types():
    mock = get_mock_data("data/get_v1_objectType_200.json")
    responses.add(
        responses.GET, mock["url"], json=mock["json"], status=mock["status_code"]
    )

    c = act.api.Act("http://localhost:8080", 1)
    object_types = c.get_object_types()

    # We should have a ipv4 object type
    assert "ipv4" in [object_t.name for object_t in object_types]

    # All object types should have a valid uuid
    assert all([re.search(UUID_MATCH, object_t.id) for object_t in object_types])


@responses.activate
def test_create_object_type():
    mock = get_mock_data("data/post_v1_objectType_threatActor_201.json")
    responses.add(
        responses.POST, mock["url"], json=mock["json"], status=mock["status_code"]
    )

    c = act.api.Act("http://localhost:8080", 1)
    object_type = c.object_type(name="threatActor", validator=".+").add()

    assert object_type.name == "threatActor"
    assert re.search(UUID_MATCH, object_type.id)


@responses.activate
def test_get_object_by_uuid():
    mock = get_mock_data("data/post_v1_object_uuid_facts_200.json")
    responses.add(
        responses.POST, mock["url"], json=mock["json"], status=mock["status_code"]
    )

    uuid = re.search(UUID, mock["url"]).group("uuid")

    c = act.api.Act("http://localhost:8080", 1)

    obj = c.object(id=uuid)

    facts = obj.facts()

    assert all([re.search(UUID_MATCH, fact.id) for fact in facts])


@responses.activate
def test_get_object_by_type_value():
    mock = get_mock_data("data/post_v1_object_facts_200.json")
    responses.add(
        responses.POST, mock["url"], json=mock["json"], status=mock["status_code"]
    )

    c = act.api.Act("http://localhost:8080", 1)

    obj = c.object(type="ipv4", value="127.0.0.1")

    facts = obj.facts()

    assert facts.complete
    assert len(facts) == facts.size

    # We should have at least one fact
    assert len(facts) >= 1

    # All facts should be of type Fact
    assert all([isinstance(fact, act.api.fact.Fact) for fact in facts])

    # All facts should have an UUID
    assert all([re.search(UUID_MATCH, fact.id) for fact in facts])


@responses.activate
def test_object_search():
    mock = get_mock_data("data/post_v1_object_search_200.json")
    responses.add(
        responses.POST, mock["url"], json=mock["json"], status=mock["status_code"]
    )

    c = act.api.Act("http://localhost:8080", 1)

    objects = c.object_search(fact_type=["seenIn"], fact_value=["report"], limit=1)

    assert not objects.complete
    assert objects.size == 1
    assert objects.count > 1

    obj = objects[0]

    # Objects should not have "object_type", since that property is stored in
    # "type"
    with pytest.raises(AttributeError):
        # pylint: disable=pointless-statement
        obj.object_type

    # Objects should not have "object_value", since that property is stored in
    # "value"
    with pytest.raises(AttributeError):
        # pylint: disable=pointless-statement
        obj.object_value

    # All facts should have an UUID
    assert all([re.search(UUID_MATCH, obj.id) for obj in objects])


@responses.activate
def test_get_object_search():
    mock = get_mock_data("data/post_v1_object_traverse_200.json")
    responses.add(
        responses.POST, mock["url"], json=mock["json"], status=mock["status_code"]
    )

    c = act.api.Act("http://localhost:8080", 1)

    obj = c.object(type="ipv4", value="127.0.0.1")

    path = obj.traverse('g.bothE("seenIn").bothV().path().unfold()')

    # Should contain both objects and facts
    assert any([isinstance(elem, act.api.obj.Object) for elem in path])
    assert any([isinstance(elem, act.api.fact.Fact) for elem in path])


def test_equality():
    c = act.api.Act("http://localhost:8080", 1)

    obj1 = c.object(type="ipv4", value="127.0.0.1")
    obj2 = c.object(
        type="ipv4", id="12345678-1234-5678-1234-567812345678", value="127.0.0.1"
    )
    obj3 = c.object(type="ipv4", value="127.0.0.2")

    assert obj1 == obj2
    assert obj1 != obj3

    assert act.api.obj.Object("fqdn") == act.api.obj.Object("fqdn")

    # Should be equal even though one of the items has an id
    assert act.api.obj.Object("fqdn") == act.api.obj.Object("fqdn", id="dummy")
    assert act.api.obj.Object("fqdn") != act.api.obj.Object("ipv4")



import copy
import functools
import ipaddress
import itertools
import os
import sys
import urllib.parse
from logging import error, warning
from typing import Iterable, List, Optional, Text, TextIO, Tuple

import act.api

from . import DEFAULT_FACT_VALIDATOR, DEFAULT_METAFACT_VALIDATOR
from .base import ActBase, Config, Origin, ValidationError
from .fact import (Fact, FactType, MetaFact, RelevantFactBindings,
                   RelevantObjectBindings, auto_fact_type)
from .obj import Object, ObjectType
from .schema import schema_doc


def as_list(value):
    "Encapsulate value in list if value is not already a list/tuple"

    if not isinstance(value, (list, tuple)):
        return [value]

    return value


def format_and_validate(facts: Iterable[Fact]) -> List[Fact]:

    fact_copies: List[Fact] = []
    for fact in facts:
        fact_copy = copy.deepcopy(fact)

        config = fact_copy.config

        if isinstance(fact_copy, Fact) and config and config.object_formatter:
            fact_copy = fact_copy.format_objects()

        if isinstance(fact_copy, Fact) and config and config.object_validator:
            try:
                fact_copy.validate_and_raise()
            except act.api.base.ValidationError as err:
                if config and config.strict_validator:
                    error(err)
                    raise
                warning(err)
                continue

        fact_copies.append(fact_copy)

    return fact_copies


def handle_facts(
    facts: Iterable[Fact],
    output_format="json",
    output_filehandle: Optional[TextIO] = None,
) -> List[Fact]:
    """

    Handle a list of facts and format and validate all facts before
    they are added to the a platform or printed to stdout

    ValidationError will cause none of the facts to be handled

    """

    if not output_filehandle:
        output_filehandle = sys.stdout

    facts = format_and_validate(facts)

    for fact in facts:
        if fact.config.act_baseurl:  # type: ignore
            fact.add()
        else:
            if output_format == "json":
                output_filehandle.write("{}\n".format(fact.json()))
            elif output_format == "str":
                output_filehandle.write("{}\n".format(str(fact)))
            else:
                raise act.api.base.ArgumentError(
                    "Illegal output_format: {}".format(output_format)
                )

    return facts


@functools.lru_cache(maxsize=4096)
def handle_fact(
    fact: Fact,
    output_format="json",
    output_filehandle: Optional[TextIO] = None,
) -> Optional[Fact]:
    """
    This function has a lru cache with size 4096, so duplicates that
    occur within this cache will be ignored.

    will use print to stdout if no file handle has been passed, otherwise
    it will write to the file handle specified
    """

    # Reuse logic from handle_facts (format, validate, add, etc)
    facts = handle_facts([fact], output_format, output_filehandle)

    # If not fact is returned, the fact failed to validate
    if not facts:
        return None

    # Return single fact
    return facts[0]


class Act(ActBase):
    """Act class exposing most of the ACT API"""

    SCHEMA = []

    def __init__(
        self,
        act_baseurl,
        user_id,
        log_level="debug",
        log_file=None,
        log_prefix="act",
        requests_common_kwargs=None,
        origin_name=None,
        origin_id=None,
        access_mode="RoleBased",
        organization=None,
        object_validator=None,
        object_formatter=None,
        strict_validator=False,
    ):
        super(Act, self).__init__()

        self.configure(
            Config(
                act_baseurl,
                user_id,
                requests_common_kwargs,
                origin_name,
                origin_id,
                access_mode,
                organization,
                object_validator,
                object_formatter,
                strict_validator,
            )
        )

        act.api.utils.setup_logging(log_level, log_file, log_prefix)

    # pylint: disable=unused-argument,dangerous-default-value
    def fact_search(
        self,
        keywords="",
        object_type=[],
        fact_type=[],
        object_value=[],
        fact_value=[],
        organization=[],
        origin=[],
        include_retracted=None,
        before=None,
        after=None,
        limit=None,
    ):
        """Search objects
        Args:
            keywords (str):               Only return Facts matching a keyword query
            object_type (str[] | str):    Only return Facts with objects having a specific
                                          ObjectType
            fact_type (str[] | str):      Only return Facts having a specific FactType
            object_value (str[] | str):   Only return Facts with objects matching a specific
                                          value
            fact_value (str[] | str):     Only return Facts matching a specific value
            organization (str[] | str):   Only return Facts belonging to
                                          a specific Organization
            origin (str[] | str):         Only return Facts coming from a specific Origin
            include_retracted (bool):     Include retracted Facts (default=False)
            before (timestamp):           Only return Facts added before a specific
                                          timestamp. Timestamp is on this format:
                                          2016-09-28T21:26:22Z
            after (timestamp):            Only return Facts added after a specific
                                          timestamp. Timestamp is on this format:
                                          2016-09-28T21:26:22Z
            limit (integer):              Limit the number of returned Objects
                                          (default 25). Limit must be <= 10000.

        All arguments are optional.

        Returns ActResultSet of Facts.
        """

        params = act.api.utils.prepare_params(
            locals(),
            ensure_list=[
                "object_type",
                "object_value",
                "fact_type",
                "fact_value",
                "organization",
                "origin",
            ],
        )

        res = self.api_post("v1/fact/search", **params)

        return act.api.base.ActResultSet(res, auto_fact_type, config=self.config)

    # pylint: disable=unused-argument,dangerous-default-value
    def object_search(
        self,
        keywords="",
        object_type=[],
        fact_type=[],
        object_value=[],
        fact_value=[],
        organization=[],
        source=[],
        before=None,
        after=None,
        limit=None,
    ):
        """Search objects
        Args:
            keywords (str):               Only return Objects matching a keyword query
            object_type (str[] | str):    Only return Objects with a specific ObjectType
            fact_type (str[] | str):      Only return Objects with Facts having a
                                          specific FactType
            object_value (str[] | str):   Only return Objects matching a specific
                                          value
            fact_value (str[] | str):     Only return Objects with Facts matching a
                                          specific value
            organization (str[] | str):   Only return Objects with Facts belonging to
                                          a specific Organization
            source (str[] | str):         Only return Objects with Facts coming from a
                                          specific Source
            before (timestamp):           Only return Objects with Facts added before
                                          a specific timestamp. Timestamp is on
                                          this format: 2016-09-28T21:26:22Z
            after (timestamp):            Only return Objects with Facts added after a
                                          specific timestamp. Timestamp is on
                                          this format: 2016-09-28T21:26:22Z
            limit (integer):              Limit the number of returned Objects
                                          (default 25, 0 means all)

        All arguments are optional.

        Returns ActResultSet of Objects.
        """

        params = act.api.utils.prepare_params(
            locals(),
            ensure_list=[
                "object_type",
                "object_value",
                "fact_type",
                "fact_value",
                "organization",
                "source",
            ],
        )

        res = self.api_post("v1/object/search", **params)

        return act.api.base.ActResultSet(res, self.object, config=self.config)

    @schema_doc(Fact.SCHEMA)
    def fact(self, *args, **kwargs):
        """Manage facts. All arguments are passed to create a Fact
        object and authentication information is passed from the
        act object."""

        f = Fact(*args, **kwargs).configure(self.config)
        f.set_defaults()

        return f

    @schema_doc(Fact.SCHEMA)
    def meta_fact(self, *args, **kwargs):
        """Manage meta facts. All arguments are passed to create a MetaFact
        object and authentication information is passed from the act object."""

        f = MetaFact(*args, **kwargs).configure(self.config)
        f.set_defaults()

        return f

    @schema_doc(Object.SCHEMA)
    def object(self, *args, **kwargs):
        """Manage objects. All arguments are passed to create an Object
        object and authentication information is passed from the
        act object."""

        return Object(*args, **kwargs).configure(self.config)

    @schema_doc(FactType.SCHEMA)
    def fact_type(self, *args, **kwargs):
        """Manage Fact Types. All arguments are passed to create a FactType
        object and authentication information is passed from the
        act object."""

        fact_type = FactType(*args, **kwargs).configure(self.config)

        if not fact_type.validator_parameter:
            fact_type.validator = "NullValidator"

            # Set to explitcit None (validator_parameter can also be empty string)
            fact_type.validator_parameter = None

        return fact_type

    @schema_doc(ObjectType.SCHEMA)
    def object_type(self, *args, **kwargs):
        """Manage Object Types. All arguments are passed to create a ObjectType
        object and authentication information is passed from the
        act object."""

        return ObjectType(*args, **kwargs).configure(self.config)

    def get_fact_types(self):
        """Get fact types"""

        return act.api.base.ActResultSet(self.api_get("v1/factType"), self.fact_type)

    def get_object_types(self):
        """Get object types"""

        return act.api.base.ActResultSet(
            self.api_get("v1/objectType"), self.object_type
        )

    @schema_doc(ObjectType.SCHEMA)
    def origin(self, *args, **kwargs):
        """Manage origins. All arguments are passed to create an origin
        object and authentication information is passed from the
        act object."""

        return Origin(*args, **kwargs).configure(self.config)

    def get_origins(self, include_deleted=False, limit=25):
        """Get origins"""

        params = {"includeDeleted": include_deleted, "limit": limit}

        return act.api.base.ActResultSet(
            self.api_get("v1/origin", params=params), self.origin
        )

    def create_fact_type(
        self,
        name,
        validator=DEFAULT_FACT_VALIDATOR,
        object_bindings=None,
        default_confidence=1.0,
    ):
        """Create fact type with given source, destination and bidirectional objects
Args:
    name (str):                  Fact type name
    validator (str):             Regular expression valdiator. Default = %s
    object_bindings (dict[]):    List of object_dict bindings
    default_confidence (float):  Default confidence for fact type

Returns created fact type, or exisiting fact type if it already exists.
""" % DEFAULT_FACT_VALIDATOR

        if not object_bindings:
            object_bindings = []

        existing_fact_types = {
            fact_type.name: fact_type for fact_type in self.get_fact_types()
        }
        object_types = {
            object_type.name: object_type for object_type in self.get_object_types()
        }

        object_types[None] = None

        # Verify that all object types exists
        for object_binding in object_bindings:
            for object_direction in ("sourceObjectType", "destinationObjectType"):
                for object_type in as_list(object_binding.get(object_direction)):
                    if object_type and object_type not in object_types:
                        raise act.api.base.ArgumentError(
                            "Object does not exist: {}".format(object_type)
                        )

        relevant_object_bindings = []

        for binding in object_bindings:
            # Default -> binding is not directional
            bidirectional = binding.get("bidirectional", False)

            source_object_type = [
                object_types[object_type]
                for object_type in as_list(binding.get("sourceObjectType", None))
            ]

            destination_object_type = [
                object_types[object_type]
                for object_type in as_list(binding.get("destinationObjectType", None))
            ]

            if not (
                "destinationObjectType" in binding or "sourceObjectType" in binding
            ):
                raise act.api.base.ArgumentError(
                    "Must specify either sourceObjectType, destinationObjectType or both in bindings for fact type {}".format(
                        name
                    )
                )

            relevant_object_bindings += [
                RelevantObjectBindings(*bindings)
                for bindings in itertools.product(
                    *[
                        as_list(source_object_type),
                        as_list(destination_object_type),
                        [bidirectional],
                    ]
                )
            ]

        if name in existing_fact_types:
            warning("Fact type %s already exists" % name)
            fact_type = existing_fact_types[name]
            fact_type.configure(self.config)
            fact_type.add_object_bindings(relevant_object_bindings)
        else:
            fact_type = self.fact_type(
                name=name,
                validator_parameter=validator,
                relevant_object_bindings=relevant_object_bindings,
                default_confidence=default_confidence,
            ).add()

        return fact_type

    def create_fact_type_all_bindings(
        self, name, validator_parameter=DEFAULT_FACT_VALIDATOR, default_confidence=1.0
    ):
        """Create a fact type that can be connected to all object types"""

        existing_fact_types = {
            fact_type.name: fact_type for fact_type in self.get_fact_types()
        }

        object_types = {
            object_type.name: object_type for object_type in self.get_object_types()
        }

        # Create list of all combiations of object types / bidirectional
        bindings = [
            RelevantObjectBindings(
                source_object_type, destination_object_type, bidirectional
            )
            for source_object_type, destination_object_type, bidirectional in itertools.product(
                *[object_types.values(), object_types.values(), [True, False]]
            )
        ]

        if name in existing_fact_types:
            # Do not create fact, but update bindings
            warning("Fact type %s already exists" % name)
            fact_type = existing_fact_types[name]
            fact_type.add_object_bindings(bindings)
        else:
            # Create fact with bindings
            fact_type = self.fact_type(
                name=name,
                validator_parameter=validator_parameter,
                relevant_object_bindings=bindings,
                default_confidence=default_confidence,
            ).add()

        return fact_type

    def create_meta_fact_type(
        self, name, fact_bindings, validator=DEFAULT_METAFACT_VALIDATOR
    ):
        """Create meta fact type with given fact bindings
Args:
    name (str):                  Fact type name
    validator (str):             Regular expression valdiator. Default = %s
    fact_bindings ([]):          List of fact bindings (name)

Returns created fact type, or exisiting fact type if it already exists.
""" % DEFAULT_METAFACT_VALIDATOR

        existing_fact_types = {
            fact_type.name: fact_type for fact_type in self.get_fact_types()
        }

        # Verify that all fact types exists
        for fact_type in fact_bindings:
            if fact_type not in existing_fact_types:
                raise act.api.base.ArgumentError(
                    "Fact type does not exist: {}".format(fact_type)
                )

        # Create list of Fact Bindings
        relevant_fact_bindings = [
            RelevantFactBindings(name=fact_type, id=existing_fact_types[fact_type].id)
            for fact_type in fact_bindings
        ]

        if name not in existing_fact_types:
            # New meta fact type
            fact_type = self.fact_type(
                name=name,
                validator_parameter=validator,
                relevant_fact_bindings=relevant_fact_bindings,
            ).add()
        else:
            # Fact type already exists. Do not create, but update bindings
            warning("Fact type %s already exists" % name)
            fact_type = existing_fact_types[name]
            fact_type.add_fact_bindings(relevant_fact_bindings)

        return fact_type

    def create_meta_fact_type_all_bindings(
        self, name, validator_parameter=DEFAULT_METAFACT_VALIDATOR
    ):
        """Create a meta fact type that can be connected to all (non-meta) fact types
Args:
    name (str):                  Fact type name
    validator (str):             Regular expression valdiator. Default = %s

Returns created fact type, or exisiting fact type if it already exists.
""" % DEFAULT_METAFACT_VALIDATOR

        # Get all existing fact types
        existing_fact_types = {
            fact_type.name: fact_type for fact_type in self.get_fact_types()
        }

        # Create list bindings for this meta fact type
        # We exclude facts that have fact bindings (meta facts)
        bindings = [
            RelevantFactBindings(name=fact.name, id=fact.id)
            for fact in existing_fact_types.values()
            if not fact.relevant_fact_bindings
        ]

        if name not in existing_fact_types:
            # New meta fact - create fact with bindings to all existing (non meta) fact types
            fact_type = self.fact_type(
                name=name,
                validator_parameter=validator_parameter,
                relevant_fact_bindings=bindings,
            ).add()

        else:
            # Fact already exists. Do not create fact, but update bindings
            warning("Meta Fact type %s already exists" % name)
            fact_type = existing_fact_types[name]
            fact_type.configure(self.config)

            # Add bindings
            fact_type.add_fact_bindings(bindings)
        return fact_type


def handle_uri(
    actapi: Act,
    uri: str,
    output_format="json",
    output_filehandle: Optional[TextIO] = None,
) -> List[Fact]:
    """Add all facts (componentOf, scheme, path, basename) from an URI to the platform

    Can raise act.api.base.ValidationError if uri does not have scheme and address component
    and strict_validator is set in config.
    Make sure to catch this exception and not create other facts to an uri that fails this
    validation as it will most likely fail later when uploading the fact to the platform.
    """

    # We do not set sys.stdout as default in the function signature
    # because that breaks redirection in pytest
    # https://github.com/pytest-dev/pytest/issues/2178

    if not output_filehandle:
        output_filehandle = sys.stdout

    facts = []

    for fact in uri_facts(actapi, uri):
        facts.append(
            handle_fact(
                fact, output_format=output_format, output_filehandle=output_filehandle
            )
        )

    return facts


def uri_facts(actapi: Act, uri: str) -> List[Fact]:
    """Get a list of all facts (componentOf, scheme, path, basename) from an URI

    Raises act.api.base.ValidationError if uri does not have scheme and address component
    and strict_validator is set in config.
    Make sure to catch this exception and not create other facts to an uri that fails this
    validation as it will most likely fail later when uploading the fact to the platform.

    Return: List of facts"""
    facts = []

    config = actapi.config

    try:
        my_uri = urllib.parse.urlparse(uri)

        scheme = my_uri.scheme
        path = my_uri.path
        query = my_uri.query
        addr = my_uri.hostname
        port = my_uri.port
    except ValueError as e:
        msg = f"Error parsing URI: {uri}: {e}"
        if config and config.strict_validator:
            error(msg)
            raise act.api.base.ValidationError(msg)
        warning(msg)
        return []

    if not (scheme and addr):
        msg = f"URI requires both scheme and address part: {uri}"
        if config and config.strict_validator:
            error(msg)
            raise act.api.base.ValidationError(msg)
        warning(msg)
        return []

    try:
        # Is address an ipv4 or ipv6?
        ip = ipaddress.ip_address(addr)
        addr_type = "ipv{}".format(ip.version)
        addr = ip.exploded
    except ValueError:
        addr_type = "fqdn"

    facts.append(
        actapi.fact("componentOf").source(addr_type, addr).destination("uri", uri)
    )

    if port:
        facts.append(actapi.fact("port", str(port)).source("uri", uri))

    facts.append(actapi.fact("scheme", scheme).source("uri", uri))

    if path and not path.strip() == "/":
        facts.append(
            actapi.fact("componentOf").source("path", path).destination("uri", uri)
        )

        basename = os.path.basename(path)

        if basename.strip():
            facts.append(actapi.fact("basename", basename).source("path", path))

    if query:
        facts.append(
            actapi.fact("componentOf").source("query", query).destination("uri", uri)
        )

    return facts


def ip_obj(addr: Text) -> Tuple[Text, Text]:
    """Return tuple of IP type and (expanded) IP address.

    Raises ValueError if addr is not valid IPv4 or IPv6 address"""

    try:
        # Is address an ipv4 or ipv6?
        ip = ipaddress.ip_address(addr.strip())

        if ip.version not in (4, 6):
            raise ValueError("Unknown IP version: %s" % ip.version)

        return ("ipv{}".format(ip.version), ip.exploded)
    except ValueError:
        raise ValueError("Invalid IP address: %s" % addr)



from logging import info, warning

import act.api

from . import DEFAULT_OBJECT_VALIDATOR
from .base import ActBase, ActResultSet, NameSpace
from .schema import Field, MissingField, schema_doc


class ObjectType(ActBase):
    SCHEMA = [
        Field("name"),
        Field("id"),
        Field("validator", default="RegexValidator"),
        Field("validator_parameter", default=DEFAULT_OBJECT_VALIDATOR),
        Field("namespace", deserializer=NameSpace),
        Field("index_option"),
    ]

    def __hash__(self):
        return hash((self.__class__.__name__, self.name, self.namespace))

    @schema_doc(SCHEMA)
    def __init__(self, *args, **kwargs):
        super(ObjectType, self).__init__(*args, **kwargs)

    def add(self):
        params = self.serialize()
        object_type = self.api_post("v1/objectType", **params)["data"]

        # Empty data and load new result from response
        self.data = {}
        self.deserialize(**object_type)

        info("Created object type: {}".format(self.name))

        return self


class ObjectStatistics(ActBase):
    """ObjectStatistics - serialized object specifying statistics"""

    SCHEMA = [
        Field("type", deserializer=ObjectType),
        Field("count"),
        Field("last_seen_timestamp", serializer=False),
        Field("last_added_timestamp", serializer=False),
    ]

    def __hash__(self):
        return hash(
            (
                self.__class__.__name__,
                self.type,
                self.count.self.last_seen_timestamp,
                self.last_added_timestamp,
            )
        )


class Object(ActBase):
    """Manage objects"""

    SCHEMA = [
        Field(
            "type",
            deserializer=ObjectType,
            serializer=lambda object_type: object_type.name,
        ),
        Field("value"),
        Field("id"),
        Field("statistics", deserializer=ObjectStatistics, serializer=False),
        Field("object", flatten=True),
        Field("direction"),
        Field("object_type", deserialize_target="type", serializer=False),
        Field("object_value", deserialize_target="value", serializer=False),
    ]

    def __hash__(self):
        return hash((self.__class__.__name__, self.type, self.value))

    @schema_doc(SCHEMA)
    def __init__(self, *args, **kwargs):
        super(Object, self).__init__(*args, **kwargs)

    def facts(self):
        """Get facts"""

        if self.id:
            url = "v1/object/uuid/{}/facts".format(self.id)
        elif self.type.name and self.value:
            url = "v1/object/{}/{}/facts".format(self.type.name, self.value)
        else:
            raise MissingField(
                "Must have either object ID or object type/value to get facts"
            )

        response = self.api_post(url)

        result_set = ActResultSet(response, act.api.fact.Fact, config=self.config)

        # Add authentication information to all facts
        return result_set("configure", self.config)

    def serialize(self):
        # Return None for empty objects (non initialized objects)
        if not (self.id or self.value):
            return None
        # return default serializer
        return super(Object, self).serialize()

    def traverse(self, query=None):
        """Traverse from object"""

        if self.id:
            url = "v1/object/uuid/{}/traverse".format(self.id)
        elif self.type.name and self.value:
            url = "v1/object/{}/{}/traverse".format(self.type.name, self.value)
        else:
            raise MissingField(
                "Must have either object ID or object type/value to get facts"
            )

        result = []
        for element in self.api_post(url, query=query)["data"]:
            if "inReferenceTo" in element:
                result.append(act.api.fact.MetaFact(**element).configure(self.config))
            elif any(["sourceObject" in element, "destinationObject" in element]):
                result.append(act.api.fact.Fact(**element).configure(self.config))
            elif "statistics" in element:
                result.append(act.api.fact.Object(**element).configure(self.config))
            else:
                warning("Unable to guess element type: {}".format(element))
                result.append(element)

        return result

    def __bool__(self):
        """Return False unless we either have an id or both type and value"""
        if self.id or (self.type and self.value):
            return True

        return False

    def __str__(self):
        """
        Construnct string representation on this format
        (src_obj_type/src_obj_value))
        """

        return "({}/{})".format(self.type.name, self.value)



import uuid

# This is a static uuid used as seed for generating reproducible uuids based on
# text This is a temporary solution until we can get sources (uuids) from the
# ACT backend
ACT_UUID_NAMESPACE = uuid.UUID("6b159598-b585-11e7-a8e7-e83935122d71")
FACT_IS_SOURCE = "FactIsSource"
FACT_IS_DESTINATION = "FactIsDestination"
ACCESS_MODES = ["Public", "RoleBased", "Explicit"]
DEFAULT_ACCESS_MODE = "RoleBased"
BIDIRECTIONAL_FACT = "BiDirectional"
DEFAULT_OBJECT_VALIDATOR = r"(.|\n)*"
DEFAULT_FACT_VALIDATOR = None
DEFAULT_METAFACT_VALIDATOR = r"(.|\n)*"
ACT_TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

from act.api import utils
from act.api import schema
from act.api import base
from act.api import obj
from act.api import fact
from act.api import helpers

from .helpers import Act



import re

import pytest
import responses
from act_test import get_mock_data

import act

# Organization is required by eval of repr(fact)
from act.api.base import Origin, Organization
from act.api.fact import Fact
from act.api.obj import Object
from act.api.re import TIMESTAMP, TIMESTAMP_MATCH, UUID, UUID_MATCH

# pylint: disable=no-member


@responses.activate
def test_add_fact():
    mock = get_mock_data("data/post_v1_fact_127.0.0.1_201.json")
    responses.add(
        responses.POST, mock["url"], json=mock["json"], status=mock["status_code"]
    )

    c = act.api.Act("http://localhost:8080", 1)

    f = (
        c.fact("seenIn", "report")
        .source("ipv4", "127.0.0.1")
        .destination("report", "xyz")
    )

    assert f.type.name == "seenIn"
    assert f.value == "report"

    # We do not have id or timestamp yet
    assert f.id is None
    assert f.timestamp is None
    assert f.source_object is not None
    assert f.destination_object is not None
    assert f.bidirectional_binding is not None

    # Add fact
    f.add()

    fact_repr = repr(f)
    repr_f = eval(fact_repr)

    assert f == repr_f
    assert f.value == repr_f.value
    assert f.type.name == repr_f.type.name
    assert f.source_object == repr_f.source_object
    assert f.source_object.value == repr_f.source_object.value
    assert f.destination_object.type == repr_f.destination_object.type
    assert f.destination_object.value == repr_f.destination_object.value

    assert str(f) == str(repr_f)

    assert (
        str(f)
        == "(ipv4/127.0.0.1) -[seenIn/report]-> (report/87428fc522803d31065e7bce3cf03fe475096631e5e07bbd7a0fde60c4cf25c7)"
    )

    # id, timestamp and organization should now be fetched from API
    assert re.search(UUID_MATCH, f.id)
    assert re.search(TIMESTAMP_MATCH, f.timestamp)
    assert re.search(UUID_MATCH, f.organization.id)
    # Not implemented/stable in backend API yet
    # self.assertRegex(f.origin.id, UUID_MATCH)


def test_add_fact_origin():
    """Tests for origin specified in config and directly on fact"""

    default_origin_name = "test-origin"
    default_origin_id = "00000000-0000-0000-0000-000000000001"

    with pytest.raises(act.api.base.ArgumentError):
        act.api.Act("", 1, origin_id="not-valid-uuid")

    # ACT client with origin name in config
    c_origin_name = act.api.Act("", 1, origin_name=default_origin_name)

    # ACT client with origin id in config
    c_origin_id = act.api.Act("", 1, origin_id=default_origin_id)

    # ACT client with not origin in config
    c_no_origin = act.api.Act("", 1)

    # Fact using origin name from config
    fact_origin_name_from_config = (
        c_origin_name.fact("seenIn", "report")
        .source("ipv4", "127.0.0.1")
        .destination("report", "xyz")
    )

    # Using origin name specified in fact
    fact_explicit_origin_name = (
        c_no_origin.fact("seenIn", "report", origin=Origin(name=default_origin_name))
        .source("ipv4", "127.0.0.1")
        .destination("report", "xyz")
    )

    # Fact using origin id from config
    fact_origin_id_from_config = (
        c_origin_id.fact("seenIn", "report")
        .source("ipv4", "127.0.0.1")
        .destination("report", "xyz")
    )

    # Using origin id specified in fact
    fact_explicit_origin_id = (
        c_no_origin.fact("seenIn", "report", origin=Origin(id=default_origin_id))
        .source("ipv4", "127.0.0.1")
        .destination("report", "xyz")
    )

    # No origin in config or in fact definition
    fact_no_origin = (
        c_no_origin.fact("seenIn", "report")
        .source("ipv4", "127.0.0.1")
        .destination("report", "xyz")
    )

    # Facts with origin name from config and explicitly defined
    # should have the same origin
    assert fact_origin_name_from_config.origin == fact_explicit_origin_name.origin

    # Facts with origin id from config and explicitly defined
    # should have the same origin
    assert fact_origin_id_from_config.origin == fact_explicit_origin_id.origin

    assert fact_no_origin.origin is None


@responses.activate
def test_add_fact_validation_error():
    """Test adding fact that fails on validation"""
    mock = get_mock_data("data/post_v1_fact_127.0.0.x_412.json")
    responses.add(
        responses.POST, mock["url"], json=mock["json"], status=mock["status_code"]
    )

    c = act.api.Act("http://localhost:8888", 1)

    f = (
        c.fact("mentions", "report")
        .source("report", "xyz")
        .destination("ipv4", "127.0.0.x")
    )

    # Add fact -> should fail on ipv4 validation
    with pytest.raises(act.api.base.ValidationError):
        f.add()

    try:
        f.add()
    except act.api.base.ValidationError as err:
        assert (
            str(err)
            == "Object did not pass validation against ObjectType. "
            + "(objectValue=127.0.0.x)"
        )


@responses.activate
def test_add_meta_fact():
    mock = get_mock_data("data/post_v1_fact_uuid_meta_201.json")
    responses.add(
        responses.POST, mock["url"], json=mock["json"], status=mock["status_code"]
    )

    c = act.api.Act("http://localhost:8080", 1)

    uuid = re.search(UUID, mock["url"]).group("uuid")

    f = c.fact(id=uuid)

    value = mock["params"]["json"]["value"]

    meta = f.meta("observationTime", value).add()

    assert meta.type.name == "observationTime"
    assert meta.value == value


@responses.activate
def test_create_fact_type():
    mock = get_mock_data("data/post_v1_factType_threatActorAlias_201.json")
    responses.add(
        responses.POST, mock["url"], json=mock["json"], status=mock["status_code"]
    )

    mock_data = mock["json"]["data"]
    threat_actor_id = mock_data["relevantObjectBindings"][0]["destinationObjectType"][
        "id"
    ]

    c = act.api.Act("http://localhost:8080", 1)
    fact_type = c.fact_type(
        name="threatActorAlias",
        validator=".+",
        relevant_object_bindings=[
            act.api.fact.RelevantObjectBindings(
                act.api.obj.Object(id=threat_actor_id),
                act.api.obj.Object(id=threat_actor_id),
                True,
            )
        ],
    ).add()

    assert fact_type.name == "threatActorAlias"
    assert re.search(UUID_MATCH, fact_type.id)


@responses.activate
def test_get_fact_types():
    mock = get_mock_data("data/get_v1_factType_200.json")
    responses.add(
        responses.GET, mock["url"], json=mock["json"], status=mock["status_code"]
    )

    c = act.api.Act("http://localhost:8080", 1)
    fact_types = c.get_fact_types()

    assert fact_types.size == len(fact_types)


@responses.activate
def test_fact_search():
    mock = get_mock_data("data/post_v1_fact_search_200.json")
    responses.add(
        responses.POST, mock["url"], json=mock["json"], status=mock["status_code"]
    )

    c = act.api.Act("http://localhost:8080", 1)

    facts = c.fact_search(fact_type=["seenIn"], fact_value=["report"], limit=1)

    assert not facts.complete
    assert facts.size == 1
    assert facts.count > 1

    # All facts should have an UUID
    assert all([re.search(UUID_MATCH, fact.id) for fact in facts])


@responses.activate
def test_fact_acl():
    mock = get_mock_data("data/get_v1_fact_uuid_access_200.json")
    responses.add(
        responses.GET, mock["url"], json=mock["json"], status=mock["status_code"]
    )

    c = act.api.Act("http://localhost:8080", 1)
    uuid = re.search(UUID, mock["url"]).group("uuid")
    acl = c.fact(id=uuid).get_acl()
    assert acl == []


@responses.activate
def test_fact_get_comments():
    mock = get_mock_data("data/get_v1_fact_uuid_comments_200.json")
    responses.add(
        responses.GET, mock["url"], json=mock["json"], status=mock["status_code"]
    )

    c = act.api.Act("http://localhost:8080", 1)

    # Get comments
    uuid = re.search(UUID, mock["url"]).group("uuid")
    comments = c.fact(id=uuid).get_comments()
    assert comments  # Should be non empty
    assert comments[0].comment == "Test comment"
    assert re.search(TIMESTAMP, comments[0].timestamp)
    assert all([isinstance(comment, act.api.base.Comment) for comment in comments])


@responses.activate
def test_fact_add_comment():
    mock = get_mock_data("data/post_v1_fact_uuid_comments_201.json")
    responses.add(
        responses.POST, mock["url"], json=mock["json"], status=mock["status_code"]
    )

    c = act.api.Act("http://localhost:8080", 1)

    # Get comments
    uuid = re.search(UUID, mock["url"]).group("uuid")

    c.fact(id=uuid).add_comment("Test comment")


def test_fact_chain_incident_organization():
    c = act.api.Act("", 1)

    facts = (
        c.fact("observedIn")
        .source("uri", "http://uri.no")
        .destination("incident", "*"),
        c.fact("targets").source("incident", "*").destination("organization", "*"),
        c.fact("memberOf").source("organization", "*").destination("sector", "energy"),
    )

    for fact in facts:
        assert any(
            [fact.source_object.value == "*", fact.destination_object.value == "*"]
        )

    seed = act.api.fact.fact_chain_seed(*facts)

    assert (
        seed
        == "(incident/*) -[targets]-> (organization/*)\n"
        + "(organization/*) -[memberOf]-> (sector/energy)\n"
        + "(uri/http://uri.no) -[observedIn]-> (incident/*)"
    )

    # Create fact chain
    chain = act.api.fact.fact_chain(*facts)

    assert len(chain) == 3

    # There should not be any placeholders any more, since these values are replaced with
    # hashes
    for fact in chain:
        assert not any(
            [fact.source_object.value == "*", fact.destination_object.value == "*"]
        )


def test_fact_chain_ta_incident():
    c = act.api.Act("", 1)

    facts = (
        c.fact("observedIn")
        .source("uri", "http://uri.no")
        .destination("incident", "*"),
        c.fact("attributedTo")
        .source("incident", "*")
        .destination("threatActor", "APT99"),
    )

    # Ensure we have placeholder objects (objects with value "*")
    for fact in facts:
        assert any(
            [fact.source_object.value == "*", fact.destination_object.value == "*"]
        )

    incident_seed = act.api.fact.fact_chain_seed(*facts)

    assert (
        incident_seed
        == "(incident/*) -[attributedTo]-> (threatActor/APT99)\n"
        + "(uri/http://uri.no) -[observedIn]-> (incident/*)"
    )

    # Create fact chain
    chain = act.api.fact.fact_chain(*facts)

    assert len(chain) == 2

    # There should not be any placeholder objects in the chain
    for fact in chain:
        assert not any(
            [fact.source_object.value == "*", fact.destination_object.value == "*"]
        )


def test_fact_chain_incident_tool():
    """Example with multiple incoming links (which should be grouped)"""
    c = act.api.Act("", 1)

    facts = (
        c.fact("observedIn")
        .source("uri", "http://uri.no")
        .destination("incident", "*"),
        c.fact("observedIn").source("tool", "mimikatz").destination("incident", "*"),
    )

    for fact in facts:
        assert any(
            [fact.source_object.value == "*", fact.destination_object.value == "*"]
        )

    incident_seed = act.api.fact.fact_chain_seed(*facts)

    assert (
        incident_seed
        == "(tool/mimikatz) -[observedIn]-> (incident/*)\n"
        + "(uri/http://uri.no) -[observedIn]-> (incident/*)"
    )


def test_fact_chain_tool_ta():
    """Example with multiple incoming links (which should be grouped)"""
    c = act.api.Act("", 1)

    facts_windshield = (
        c.fact("attributedTo")
        .source("incident", "*")
        .destination("threatActor", "APT32"),
        c.fact("observedIn", "incident")
        .source("content", "*")
        .destination("incident", "*"),
        c.fact("classifiedAs").source("content", "*").destination("tool", "windshield"),
    )

    facts_mimikatz = (
        c.fact("attributedTo")
        .source("incident", "*")
        .destination("threatActor", "APT32"),
        c.fact("observedIn", "incident")
        .source("content", "*")
        .destination("incident", "*"),
        c.fact("classifiedAs").source("content", "*").destination("tool", "mimikatz"),
    )

    windshield_seed = act.api.fact.fact_chain_seed(*facts_windshield)
    mimikatz_seed = act.api.fact.fact_chain_seed(*facts_mimikatz)

    assert (
        windshield_seed
        == "(content/*) -[classifiedAs]-> (tool/windshield)\n"
        + "(content/*) -[observedIn/incident]-> (incident/*)\n"
        + "(incident/*) -[attributedTo]-> (threatActor/APT32)"
    )

    assert windshield_seed != mimikatz_seed


def test_fact_chain_illegal_tool():
    """Example with multiple incoming links (which should be grouped)"""
    c = act.api.Act("", 1)

    facts = (
        c.fact("observedIn")
        .source("uri", "http://uri.no")
        .destination("incident", "my-incident"),
        c.fact("observedIn").source("tool", "mimikatz").destination("incident", "*"),
    )

    # Create fact chain
    with pytest.raises(act.api.fact.IllegalFactChain):
        act.api.fact.fact_chain(*facts)


def test_fact_equality():
    c = act.api.Act("http://localhost:8080", 1)
    c2 = act.api.Act("http://localhost:8080", 1, origin_name="dummy")

    f1 = (
        c.fact("observedIn")
        .source("uri", "http://uri.no")
        .destination("incident", "my-incident")
    )
    f2 = (
        c2.fact("observedIn")
        .source("uri", "http://uri.no")
        .destination("incident", "my-incident")
    )
    f3 = (
        c.fact("observedIn", origin="dummy")
        .source("uri", "http://uri.no")
        .destination("incident", "my-incident")
    )

    # Hash of two facts with different origin should not be equal
    assert hash(f1) != hash(f2)
    assert f1 != f2

    # Hash of facts created with origin from default and explitcit added should be equal
    assert hash(f2) == hash(f3)
    assert f2 == f3

    m1 = f1.meta("observationTime", "1624450340")

    # The hash of the `referenced` fact in a meta fact should be equal to the hash
    # of the fact that was used to create the meta fact
    assert hash(m1.in_reference_to) == hash(f1)
    assert m1.in_reference_to == f1

    assert act.api.fact.FactType("observedIn") == act.api.fact.FactType("observedIn")
    assert act.api.fact.FactType("observedIn") == act.api.fact.FactType(
        "observedIn", id="dummy"
    )



import re

import responses
from act_test import get_mock_data

import act
from act.api.re import UUID_MATCH


# pylint: disable=no-member
@responses.activate
def test_get_origins():
    mock = get_mock_data("data/get_v1_origin_200.json")
    responses.add(
        responses.GET, mock["url"], json=mock["json"], status=mock["status_code"]
    )

    c = act.api.Act("http://localhost:8888", 1)

    origins = c.get_origins(limit=0)

    # We should have a ipv4 object type
    assert "my-origin" in [origin.name for origin in origins]

    # All origins should have a valid uuid
    assert all([re.search(UUID_MATCH, origin.id) for origin in origins])


@responses.activate
def test_create_origin():
    mock_add = get_mock_data("data/post_v1_origin_myorigin_201.json")
    responses.add(
        responses.POST,
        mock_add["url"],
        json=mock_add["json"],
        status=mock_add["status_code"],
    )

    mock_get = get_mock_data("data/get_v1_origin_200.json")
    responses.add(
        responses.GET,
        mock_get["url"],
        json=mock_get["json"],
        status=mock_get["status_code"],
    )

    c = act.api.Act("http://localhost:8888", 1)

    o = c.origin("my-origin", trust=0.8, description="My origin")
    origin = o.add()

    assert origin.name == "my-origin"
    assert re.search(UUID_MATCH, origin.id)



import pytest

from act.api.base import ActBase, Comment, NameSpace, Organization, Origin


class Child(ActBase):
    pass


def test_equailty():

    with pytest.raises(NotImplementedError):
        assert Child() == Child()

    ns_mnemonic = NameSpace("mnemonic")
    ns_google = NameSpace("google")
    org_mnemonic = Organization("mnemonic")
    org_google = Organization("google")
    origin_mnemonic = Origin(
        "mnemonic", namespace=ns_mnemonic, organization=org_mnemonic
    )
    origin_google = Origin("google", namespace=ns_google, organization=org_google)

    assert ns_mnemonic == NameSpace("mnemonic", "dummy-id")
    assert ns_mnemonic != ns_google
    assert org_mnemonic == Organization("mnemonic", "dummy-id")
    assert org_mnemonic != org_google

    assert origin_mnemonic == Origin(
        "mnemonic", "dummy-id", namespace=ns_mnemonic, organization=org_mnemonic
    )

    assert origin_mnemonic != origin_google

    assert Comment("a") == Comment("a")
